package fi.jyu.mit.graphics;

import java.awt.Color;
import java.awt.Graphics;

/**
 * Kaikkien kuvioiden yliluokka
 * @author Markus Kivioja
 * @author Vesa Lappalainen
 */
public abstract class BasicShape implements Drawable { // NOPMD

    private Color color;
    private Matrix matrix;
    private Matrix rotator;
    private BasicDrawableCollection parent;


    public BasicShape() {
        this.matrix = null; // NOPMD
        this.rotator = null;// NOPMD
        this.color = null; // NOPMD
    }


    @Override
    public Drawable setColor(int r, int g, int b) { // NOPMD
        return setColor(new Color(r, g, b));
    }


    /**
     * Asetetaan piirtoobjecktin vri
     * @see fi.jyu.mit.graphics.Drawable#setColor(java.awt.Color)
     */
    @Override
    public Drawable setColor(Color c) { // NOPMD
        this.color = c;
        redraw();
        return this;
    }


    /**
     * Palautetaan vri jota kyten piirtmiseen
     * @return kuvion vri
     */
    @Override
    public Color getColor() {
        if (color != null)
            return this.color;
        if (parent == null) {
            return Color.BLACK;
        }
        return parent.getColor();
    }


    /**
     * Alustetaan graphics nykyisten tietojen pohjalta.
     * Vari haetaan omistajalta jos itsella ei ole.
     * Jos kenellakaan ei ole varia, kaytetaan mustaa.
     * @param g Graphics joka alustetaan
     */
    protected void initGraphics(Graphics g) {
        if (color == null) {
            if (parent == null) {
                g.setColor(Color.BLACK);
                return;
            }
            parent.initGraphics(g);
            return;
        }
        g.setColor(this.color);
    }


    /**
     * Palauttaa annetun matriisin muunnettuna kuvion omalla muunnosmatriisilla,
     * jos muunnosmatriisia ei ole asetettu se on yksikkmatriisi
     * @param a muunnettava matriisi
     * @return a muunnettuna kuvion omalla matriisilla
     */
    public Matrix getTransform(Matrix a) { // NOPMD
        if (matrix == null)
            return a;
        if (a == null)
            return getTransform();
        return a.multiply(matrix);
    }


    /**
     * Palauttaa annetun matriisin kierrettyn kuvion omalla kiertomatriisilla,
     * jos kiertomatriisia ei ole asetettu se on yksikkmatriisi
     * @param a muunnettava matriisi
     * @return a muunnettuna kuvion omalla matriisilla
     */
    public Matrix getRotator(Matrix a) { // NOPMD
        if (rotator == null)
            return a;
        if (a == null)
            return rotator;
        return a.multiply(rotator);
    }


    @Override
    public Matrix getTransform() { // NOPMD
        if (matrix == null)
            matrix = new Matrix();
        return matrix;
    }


    /**
     * Muuntaa annetun pisteen annetulla matriisilla ja kuvion omalla muunnosmatriisilla,
     * palauttaa muunnetun pisteen SPoint-oliona
     * @param a muunnosmatriisi
     * @param x muunnettavan pisteen x-koordinaatti
     * @param y muunnettavan pisteen y-koordinaatti
     * @param z muunnettavan pisteen z-koortinaatti
     * @return muunnettu piste SPoint-oliona
     */
    public SPoint getPoint(Matrix a, double x, double y, double z) { // NOPMD
        Matrix tr = getTransform(a);
        if (tr != null)
            return tr.transform(x, y, z);
        return new SPoint((int) x, (int) y);
    }


    /**
     * Pirt kuvion annetulla Graphics-oliolla ja muunnosmatriisilla
     * @param g Graphics-olio
     * @param a muunnosmatriisi
     */
    abstract protected void drawShape(Graphics g, Matrix a);


    @Override
    public void draw(Graphics g, Matrix a) {
        Matrix t = getRotator(getTransform(a));
        initGraphics(g);
        drawShape(g, t);
    }


    @Override
    public void setParent(BasicDrawableCollection parent) {
        if (this.parent == parent)
            return;
        if (this == parent)
            return; // itseaan ei saa lisata itseensa
        if (this.parent != null)
            this.parent.justRemove(this);
        this.parent = parent;
        if (parent != null)
            parent.justAdd(this);
        redraw();
    }


    @Override
    public void redraw() {
        if (this.parent == null)
            return;
        this.parent.redraw();
    }


    @Override
    public Drawable transform(Matrix m) { // NOPMD
        if (this.matrix == null)
            this.matrix = new Matrix();
        matrix.multiplyThis(m);
        redraw();
        return this;
    }


    @Override
    public Drawable setTransform(Matrix m) { // NOPMD
        if (this.matrix == m)
            return this;
        this.matrix = m;
        redraw();
        return this;
    }


    @Override
    public void changeTransform(Matrix m) { // NOPMD
        setTransform(m);
    }


    @Override
    public void setRotator(Matrix m) { // NOPMD
        this.rotator = m;
        redraw();
    }


    @Override
    public Matrix getRotator() { // NOPMD
        if (this.rotator == null)
            this.rotator = new Matrix();
        return this.rotator;
    }


    @Override
    public Drawable scale(double dx, double dy, double dz) { // NOPMD
        return transform(new ScaleMatrix(dx, dy, dz));
    }


    @Override
    public Drawable move(double sx, double sy, double sz) { // NOPMD
        return transform(new TranslateMatrix(sx, sy, sz));
    }


    @Override
    public Drawable rotate(int axis, double deg) { // NOPMD
        return transform(new RotMatrix(axis, deg));
    }

}
