package fi.jyu.mit.graphics;

import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Hashtable;

import javax.imageio.ImageIO;

/**
 * Yksinkertainen bitmapin piirto.
 * Muunnosmatriisit vaikuttavat ainoastaan paikkaan (toistaiseksi).
 * Bitmapista voidaan ottaa rakadata ja tehd sille hienompia temppuja
 * @see BufferedImage 
 * @author vesal
 * @version 4.9.2010
 */
public class Bitmap extends BasicShape {

    private final RPoint upperLeft;
    int imgWidth = 0;
    int imgHeight = 0;

    private BufferedImage rawImg;


    /**
     * Luodaan bitmap tiedostosta tai URLista jos nimi alkaa http.
     * Jos tiedostoa ei ole, kokeillaan resursseista
     * @param x koordinaatti johon bitmap nytetn
     * @param y koordinaatti johon bitmap nytetn
     * @param z syvyyskoordinaatti johon bitmap nytetn
     * @param filename tiedoston nimi, josta bitmap luetaan 
     */
    public Bitmap(double x, double y, double z, String filename) {
        super();
        upperLeft = new RPoint(x, y, z);
        try {
            if (filename.startsWith("http"))
                rawImg = ImageIO.read(new URL(filename));
            else {
                File file = new File(filename);
                if (file.exists())
                    rawImg = ImageIO.read(file);
                else
                    rawImg = ImageIO.read(getClass().getResource(
                            (filename.startsWith("/") ? "" : "/") + filename));
            }
        } catch (Exception e) {
            // e.printStackTrace();
            throw new NullPointerException(
                    "filename " + filename + " not found");
        }
        imgWidth = rawImg.getWidth(null);
        imgHeight = rawImg.getHeight(null);
    }


    /**
     * Luodaan bitmap tietovirrasta
     * @param x koordinaatti johon bitmap nytetn
     * @param y koordinaatti johon bitmap nytetn
     * @param z syvyyskoordinaatti johon bitmap nytetn
     * @param stream tietovirta, josta luetaan 
     */
    public Bitmap(double x, double y, double z, InputStream stream) {
        super();
        upperLeft = new RPoint(x, y, z);
        try {
            rawImg = ImageIO.read(stream);
        } catch (Exception e) {
            // e.printStackTrace();
            throw new NullPointerException("problems while reading image");
        } finally {
            try {
                stream.close();
            } catch (IOException e) {
                //
            }
        }
        imgWidth = rawImg.getWidth(null);
        imgHeight = rawImg.getHeight(null);
    }


    /**
     * Luodaan bitmap tiedostosta
     * @param x koordinaatti johon bitmap nytetn
     * @param y koordinaatti johon bitmap nytetn
     * @param filename tiedoston nimi, josta bitmap luetaan 
     */
    public Bitmap(double x, double y, String filename) {
        this(x, y, 0, filename);
    }


    /**
     * Luodaan tyhj valkoinen bitmap 
     * @param x koordinaatti johon bitmap nytetn
     * @param y koordinaatti johon bitmap nytetn
     * @param z syvyyskoordinaatti johon bitmap nytetn
     * @param w uuden kuvan leveys 
     * @param h uuden kuvan korkeus
     */
    public Bitmap(double x, double y, double z, int w, int h) {
        super();
        upperLeft = new RPoint(x, y, z);
        rawImg = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
        imgWidth = rawImg.getWidth(null);
        imgHeight = rawImg.getHeight(null);
    }


    /**
     * Luodaan tyhj valkoinen bitmap 
     * @param x koordinaatti johon bitmap nytetn
     * @param y koordinaatti johon bitmap nytetn
     * @param w uuden kuvan leveys 
     * @param h uuden kuvan korkeus
     */
    public Bitmap(double x, double y, int w, int h) {
        this(x, y, 0, w, h);
    }


    /**
     * Apumetodi kopion tekemiseksi raakasta kuvasta.
     * @param image kuvatieto joka kopioidaan
     * @return kopioitu kuvatieto
     */
    public static BufferedImage cloneBufferedImage(BufferedImage image) {
        String[] pnames = image.getPropertyNames();
        Hashtable<String, Object> cproperties = new Hashtable<String, Object>();
        if (pnames != null) {
            for (int i = 0; i < pnames.length; i++) {
                cproperties.put(pnames[i], image.getProperty(pnames[i]));
            }
        }
        WritableRaster wr = image.getRaster();
        WritableRaster cwr = wr.createCompatibleWritableRaster();
        cwr.setRect(wr);
        BufferedImage cimage = new BufferedImage(image.getColorModel(), cwr,
                image.isAlphaPremultiplied(), cproperties);
        return cimage;
    }


    /**
     * Luodaan uusi Bitmap vanhan aidoksi kopioiksi
     * @param x koordinaatti johon bitmap nytetn
     * @param y koordinaatti johon bitmap nytetn
     * @param z syvyyskoordinaatti johon bitmap nytetn
     * @param bm kopioitava kuva
     */
    public Bitmap(double x, double y, double z, Bitmap bm) {
        super();
        upperLeft = new RPoint(x, y, z);
        rawImg = cloneBufferedImage(bm.getRawImage());
        imgWidth = rawImg.getWidth(null);
        imgHeight = rawImg.getHeight(null);
    }


    /**
     * Luodaan uusi Bitmap vanhan aidoksi kopioiksi
     * @param x koordinaatti johon bitmap nytetn
     * @param y koordinaatti johon bitmap nytetn
     * @param bm kopioitava kuva
     */
    public Bitmap(double x, double y, Bitmap bm) {
        this(x, y, 0, bm);
    }


    /**
     * @param p pikseli jonka kirkkaus lasketaan
     * @return pikselin kirkkaus arvona 0-255
     */
    public static int brightness(int p) {
        int r = p & 0xff;
        p >>= 8;
        r += p & 0xff;
        p >>= 8;
        r += p & 0xff;
        return r / 3;
    }


    /**
     * @return kuvan leveys
     */
    public int getWidth() {
        return imgWidth;
    }


    /**
     * @return kuvan korkeus
     */
    public int getHeight() {
        return imgHeight;
    }


    /**
     * Palautetaan kuvan yhden pikselin vritiedot
     * @param x pikselin x-koordinaatti
     * @param y pikselin y-koordinaatti
     * @return RGB ja alpha tieto pikselist
     */
    public int getRGB(int x, int y) {
        return rawImg.getRGB(x, y);
    }


    /**
     * Asetetaan pikselin RGB ja alpha arvot
     * @param x pikselin x-koordinaatti
     * @param y pikselin y-koordinaatti
     * @param rgb uusi RGB ja alpha arvo pikselille
     */
    public void setRGB(int x, int y, int rgb) {
        rawImg.setRGB(x, y, rgb);
    }


    /**
     * Muuttaa kolme vri ja alpha-tiedon yhdeksi vritiedoksi
     * @param r punaisen mr (0-255)
     * @param g vihren mr (0-255)
     * @param b sinisen mr (0-255)
     * @param a alha-arvo
     * @return tieto pakattuna yhteen 32-bit lukuun
     */
    public static int toRGB(int r, int g, int b, int a) {
        return (a << 24) | (r << 16) | (g << 8) | b;
    }


    /**
     * Muuttaa kolme vri 
     * @param r punaisen mr (0-255)
     * @param g vihren mr (0-255)
     * @param b sinisen mr (0-255)
     * @return tieto pakattuna yhteen 32-bit lukuun
     */
    public static int packRGB(int r, int g, int b) {
        return (255 << 24) | (r << 16) | (g << 8) | b;
    }


    /**
     * @param c vri jota tutkitaan
     * @return vrist otettu punaisen mr (0-255)
     */
    public static int getRed(int c) {
        return (c >> 16) & 0xff;
    }


    /**
     * @param c vri jota tutkitaan
     * @return vrist otettu vihren mr (0-255)
     */
    public static int getGreen(int c) {
        return (c >> 8) & 0xff;
    }


    /**
     * @param c vri jota tutkitaan
     * @return vrist otettu sinisen mr (0-255)
     */
    public static int getBlue(int c) {
        return (c >> 0) & 0xff;
    }


    /**
     * @param c vri jota tutkitaan
     * @return vrist otettu alpha-arvo
     */
    public static int getA(int c) {
        return (c >> 24) & 0xff;
    }


    /**
     * @return viite raakaan kuvadata monimutkaisempaa ksittely varten
     * @see BufferedImage
     */
    public BufferedImage getRawImage() {
        return rawImg;
    }


    @Override
    protected void drawShape(Graphics g, Matrix a) {
        SPoint sp = new SPoint(0, 0);
        a.transform(upperLeft, sp);
        g.drawImage(rawImg, sp.getX(), sp.getY(), null);
    }


    /**
     * Muuttaa kuvan harmaasvyiksi
     */
    public void convertGrayScale() {
        for (int y = 0; y < rawImg.getHeight(); y++) {
            for (int x = 0; x < rawImg.getWidth(); x++) {
                int c = rawImg.getRGB(x, y);
                int b = Bitmap.brightness(c);
                b = (b << 16) | (b << 8) | b;
                rawImg.setRGB(x, y, b);
            }
        }
    }


    /**
     * Tekee jokaiselle pikselille and operaation maskin kanssa.
     * @param mask maski, jolla pikseli muutetaan
     */
    public void andPixels(int mask) {
        for (int y = 0; y < rawImg.getHeight(); y++) {
            for (int x = 0; x < rawImg.getWidth(); x++) {
                int c = rawImg.getRGB(x, y);
                rawImg.setRGB(x, y, c & mask);
            }
        }
    }


    /**
     * Tekee jokaiselle pikselille OR operaation maskin kanssa.
     * @param mask maski, jolla pikseli muutetaan
     */
    public void orPixels(int mask) {
        for (int y = 0; y < rawImg.getHeight(); y++) {
            for (int x = 0; x < rawImg.getWidth(); x++) {
                int c = rawImg.getRGB(x, y);
                rawImg.setRGB(x, y, c | mask);
            }
        }
    }


    /**
     * Luodaan kuvasta 2-ulotteinen taulukko, johon kopioidaan kuvan pisteet 
     * @param ox   - siirtym x-suunnassa vasemmasta ylnurkasta
     * @param oy   - siirtym y-suunnassa vasemmasta ylnurkasta
     * @param w    - alueen leveys josta kopioidaan
     * @param h    - alueen korkeus josta kopioidaan
     * @return kuvan bittitieto 2.ulotteisena kokonaislukutaulukkona
     */
    public int[][] getData(int ox, int oy, int w, int h) {
        int maxx = rawImg.getWidth() - ox;
        int maxy = rawImg.getHeight() - oy;
        if (ox + w < maxx)
            maxx = w;
        if (oy + h < maxy)
            maxy = h;

        int[][] data = new int[maxy][maxx];
        for (int y = 0; y < maxy; y++)
            for (int x = 0; x < maxx; x++)
                data[y][x] = rawImg.getRGB(x + ox, y + oy);
        return data;
    }


    /**
     * @return kuvan bittitieto 2.ulotteisena kokonaislukutaulukkona
     */
    public int[][] getData() {
        return getData(0, 0, rawImg.getWidth(), rawImg.getHeight());
    }


    /**
     * Asettaa bittidatantaulukon haluttuun kohtaan kuvassa.
     * Jos kuva menee ulos , ei ulos menevi pisteit huomioida.
     * Taulukossa saa olla erimittaisia rivej
     * 
     * @param ox   - siirtym x-suunnassa vasemmasta ylnurkasta
     * @param oy   - siirtym y-suunnassa vasemmasta ylnurkasta
     * @param w    - alueen leveys johon kopioidaan
     * @param h    - alueen korkeus johon kopioidaan
     * @param data - kopioitava taulukko
     */
    public void setData(int ox, int oy, int w, int h, int[][] data) {
        // int[][] data = new int[rawImg.getHeight()][rawImg.getWidth()];
        int imgw = rawImg.getWidth();
        int imgh = rawImg.getHeight();
        int maxy = data.length;
        if (h < maxy)
            maxy = h;
        if (imgh < maxy + oy)
            maxy = imgh - oy;
        for (int y = 0; y < maxy; y++) {
            int maxx = data[y].length;
            if (w < maxx)
                maxx = w;
            if (imgw < maxx + ox)
                maxx = imgw - ox;
            for (int x = 0; x < maxx; x++)
                rawImg.setRGB(ox + x, oy + y, data[y][x]);
        }
    }


    /**
     * Asettaa bittidatantaulukon haluttuun kohtaan kuvassa.
     * Jos kuva menee ulos , ei ulos menevi pisteit huomioida.
     * Taulukossa saa olla erimittaisia rivej, 1. rivi mr leveyden
     * 
     * @param ox   - siirtym x-suunnassa vasemmasta ylnurkasta
     * @param oy   - siirtym y-suunnassa vasemmasta ylnurkasta
     * @param data - kopioitava taulukko
     */
    public void setData(int ox, int oy, int[][] data) {
        setData(ox, oy, data[0].length, data.length, data);
    }

}
