package fi.jyu.mit.graphics;

import java.awt.*;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;

/**
 * Ikkunoissa nytettv paneeli johon kuviot piirretn
 * @author Markus Kivioja
 * @author Vesa Lappalainen
 *
 */
public class SpacePanel extends Component implements SpaceParent {

    private static final long serialVersionUID = 1L;
    private Matrix sm; // Perusskaalaus kuva ikkunaan
    private Matrix a; // Muunnokset taman paalle
    private Matrix tm; // Kokonaismuunnos tm = sm * a;
    private Matrix rotator;
    private boolean isotrophic;
    private boolean allowRescale = true;
    private Space space;
    private RRect rRect = null;
    private double left, right, top, bottom;


    /**
     * Luo uuden paneelin jonka taustavri on valkoinen
     */
    public SpacePanel() {
        super();
        initialize(); // NOPMD
        setBackground(Color.WHITE);
        // setOpaque(false);
        setBackground(Color.WHITE);
    }


    /**
     * Alustaa paneelin asetuksen, kertoo nyttmlleen avaruudelle,
     * ett tm paneeli nytt sit
     */
    private final void initialize() {
        this.space = new Space();
        this.space.registerWindow(this);
        this.sm = new TMatrix();
        this.a = new TMatrix();
        this.tm = new TMatrix();
        this.rotator = new TMatrix();
        // this.height = getHeight();
        this.isotrophic = true;
        addComponentListener(new ComponentAdapter() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void componentResized(ComponentEvent e) {
                if (allowRescale)
                    reScale(); // NOPMD
            }
        });
    }


    /**
     * Skaalaa paneelin sislln annettujen reuna-arvojen mukaisesti, 
     * esimerkiksi:
     * <pre>
     *     panel.scale(-2.5, -2.2, 2.1, 2.7);
     * </pre>
     * @param left sislln x-koordinaatin arvo paneelin vasemmassa reunassa
     * @param bottom sislln y-koordinaatin arvo paneelin alareunassa 
     * @param right sislln x-koordinaatin arvo paneelin oikeassa reunassa
     * @param top sislln y-koordinaatin arvo paneelin ylreunassa
     */
    public void scale(double left, double bottom, double right, double top) {
        this.left = left;
        this.bottom = bottom;
        this.right = right;
        this.top = top;
        rRect = new RRect(left, top, right, bottom);
        reScale();
    }


    /**
     * Skaalaa paneelin sislln oikean kokoiseksi jos paneelin kokoa muutetaan
     */
    public final void reScale() {
        if (rRect == null)
            return;
        SRect sRect = new SRect(0, 0, getWidth(), getHeight());
        this.sm = new TMatrix(rRect, sRect, isotrophic);
        redraw();
    }


    /**
     * Pyritt paneelin sislt annetun akselin ympri, esimerkiksi:
     * <pre>
     *     panel.rotate(Axis.X, 45.2);
     * </pre>
     * @param axis akseli jonka ympri pyritetn (Axis.X/Axis.Y/Axis.Z)
     * @param deg pyritettvn kulman suuruus asteina
     */
    public void rotate(int axis, double deg) {
        transform(new RotMatrix(axis, deg));
    }


    /**
     * Liikuttaa paneelin sislt, esimerkiksi:
     * <pre>
     *     panel.move(0, 1.5, -15);
     * </pre>
     * @param dx liikutettava matka x-akselin suunnassa
     * @param dy liikutettava matka y-akselin suunnassa
     * @param dz liikutettava matka z-akselin suunnassa
     */
    public void move(double dx, double dy, double dz) {
        transform(new TranslateMatrix(dx, dy, dz));
    }


    /**
     * Skaalaa paneelin sislln annettujen kertoimien mukaan
     * <pre>
     *      panel.scale(2, 1, 1); venytt kuviot kaksinkertaiseksi x-suunnassa
     * </pre>
     * @param sx sislln x-suuntainen kerroin
     * @param sy sislln y-suuntainen kerroin
     * @param sz sislln z-suuntainen kerroin
     */
    public void scale(double sx, double sy, double sz) {
        this.left = this.left * sx;
        this.right = this.right * sx;
        this.bottom = this.bottom * sy;
        this.top = this.top * sy;
        transform(new ScaleMatrix(sx, sy, sz));
    }


    /**
     * @see fi.jyu.mit.graphics.Transformable#transform(fi.jyu.mit.graphics.Matrix)
     */
    @Override
    public Transformable transform(Matrix m) {
        if (m == null)
            return this;
        this.a.multiplyThis(m);
        redraw();
        return this;
    }


    /**
     * Asetetaan kirtomatriisi
     * @param m
     */
    public void setTransform(Matrix m) {
        this.a = m;
        redraw();
    }


    /**
     * @see fi.jyu.mit.graphics.Transformable#changeTransform(fi.jyu.mit.graphics.Matrix)
     */
    @Override
    public void changeTransform(Matrix m) {
        setTransform(m);
    }


    /**
     * @see fi.jyu.mit.graphics.Transformable#setRotator(fi.jyu.mit.graphics.Matrix)
     */
    @Override
    public void setRotator(Matrix m) {
        this.rotator = m;
        redraw();
    }


    /**
     * @see fi.jyu.mit.graphics.Transformable#getRotator()
     */
    @Override
    public Matrix getRotator() {
        return this.rotator;
    }


    /**
     * Asettaa paneelin nyttmn annettua avaruutta
     * @param space avaruus jota nytetn
     */
    public void setSpace(Space space) {
        this.space = space;
        this.space.registerWindow(this);
    }


    /**
     * @return panelin kyttm avaruus
     */
    public Space getSpace() {
        return space;
    }


    /**
     * Lis paneeliin kuvion joka on Drawable-olio ja palauttaa sen,
     * esimerkiksi:
     * <pre>
     *      panel.add(new Line(0, 0, 1, 1).setColor(255, 255, 0));
     * </pre>
     * @param shape listtv kuvio
     * @return listty kuvio
     */
    public Drawable add(Drawable shape) {
        return this.space.add(shape);
    }


    /**
     * Antaa Drawable-olion joka muodostuu kaikista 
     * tmn paneelin nyttmist kuvioista
     * @return kokoelma tmn paneelin nyttmist kuvioista
     */
    public Drawable getSavedPath() {
        return this.space.getSavedPath();
    }


    /**
     * Poistaa paneelista kuvion joka on Drawable-olio
     * @param shape poistettava kuvio
     */
    public void remove(Drawable shape) {
        this.space.remove(shape);
    }


    public void removeAll() {
        this.space.removeAll();
    }


    /**
     * Piirt paneelin sislln uudelleen
     */
    @Override
    public void redraw() {
        repaint(1000L);
    }


    /**
     * Antaa paneelin kokonaismuunnosmatriisin
     * @return paneelin kokonaismuunnosmatriisi
     */
    public Matrix getFullTransform() {
        return tm;
    }


    /**
     * Piirt paneelin sislln omalla vrilln ja muunnosmatriiseillaan
     * @param g grafiikka johon piirretn
     */
    @Override
    public void paint(Graphics g) {
        g.setColor(getBackground());
        g.fillRect(0, 0, getWidth(), getHeight());
        if (space == null)
            return;
        sm.multiply(a, tm);
        tm.multiplyThis(rotator);
        this.space.draw(g, tm);
    }


    /**
     * Asettaa paneelin nkyvksi 
     */
    public void showWindow() {
        this.setVisible(true);
    }


    /**
     * Asettaa isotrooppisuuden plle/pois
     * @param isotrophic isotrooppisuus tosi/eptosi
     */
    public void setIsotrophic(boolean isotrophic) {
        this.isotrophic = isotrophic;
    }


    /**
     * Asettaa uudelleenskaalauksen kun paneelin kokoa muutetaan plle/pois
     * @param allowRescale uudelleenskaalaus tosi/eptosi
     */
    public void setAllowRescale(boolean allowRescale) {
        this.allowRescale = allowRescale;
        if (allowRescale)
            reScale();
    }


    /**
     * Antaa paneelin muunnosmatriisin
     * @return paneelin muunnosmatriisi
     */
    @Override
    public Matrix getTransform() {
        return a;
    }


    public double getLeft() {
        return this.left;
    }


    public double getRight() {
        return this.right;
    }


    public double getTop() {
        return this.top;
    }


    public double getBottom() {
        return this.bottom;
    }
}
