package fi.jyu.mit.graphics;

import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.JFrame;

/**
 * Window-luokka, ikkuna joka nytt avaruudet.
 * Ikkunaan voi list kuvioita, esimerkiksi:
 * esimerkiksi:
 * <pre>
 *      window.add(new Line(0, 0, 1, 1)).setColor(255, 255, 0);
 * </pre> 
 * Ikkunaa voi skaalata, kiert ja liikuttaa, esimerkiksi:
 * <pre>
 *      window.scale(-2.5, -2.2, 2.1, 2.7);
 *      window.rotate(Axis.X, 45.2);
 *      window.move(0, 1.5, -15);
 * </pre>
 * @author Markus Kivioja
 */
public class Window extends JFrame implements Transformable {

    private static final long serialVersionUID = 1L;
    private SpacePanel panel;
    private ControllerFrame touchPad, controlButtons;
    private static Window lastOpenedWindow = null;


    /**
     * @return palauttaa viimeksi luodun ikkunan 
     */
    public static Window getLastOpenedWindow() {
        return lastOpenedWindow;
    }


    /**
     * Capture frame
     * @param screenRect area to capture
     * @param filename where to capture
     */
    public static void capture(Rectangle screenRect, String filename) {
        BufferedImage capture;
        try {
            capture = new Robot().createScreenCapture(screenRect);
            ImageIO.write(capture, "png", new File(filename));
        } catch (AWTException e) {
            //
        } catch (IOException e) {
            //
        }
    }


    /**
     * Capture screen
     * @param filename where to capture
     */
    public static void capture(String filename) {
        capture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize()),
                filename);
    }


    /**
     * Use this method to create a screenshot of the JFrame object argFrame.
     *
     * Author(s):
     *   Dejan Lekic
     * License: 
     *   Public Domain
     *
     * @param frame you want to make screenshot of.
     * @param filename where to capture
     */
    public static final void capture(Container frame, String filename) {
        Rectangle rec = frame.getBounds();
        BufferedImage bufferedImage = new BufferedImage(rec.width, rec.height,
                BufferedImage.TYPE_INT_ARGB);
        frame.paint(bufferedImage.getGraphics());

        try {
            ImageIO.write(bufferedImage, "png", new File(filename));
        } catch (IOException ioe) {
            //
        }
    }


    /**
     * Luo uuden ikkunan jonka korkeus on 500 ja leveys 500
     *
     */
    public Window() {
        this(500, 500);
    }


    /**
     * Luo uuden ikkunan
     * @param width leveys
     * @param height korkeus
     */
    public Window(int width, int height) {
        super();
        this.setSize(width, height);
        initialize();
        setBackground(Color.WHITE);
    }


    /**
     * Alustaa uuden ikkunan asetukset
     */
    private void initialize() {
        Container contentPane = this.getContentPane();
        contentPane.setLayout(new BorderLayout());
        panel = new SpacePanel();
        panel.setSize(getWidth(), getHeight());
        contentPane.add(panel);
        addWindowListener(new CloseWindow());
        lastOpenedWindow = this;

        final String captureName = System.getenv("captureName");
        if (captureName == null)
            return;

        final Frame wnd = this;
        EventQueue.invokeLater(new Runnable() {
            @Override
            public void run() {
                try {
                    Thread.sleep(1000);
                    capture(wnd, captureName);
                    System.exit(0);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        });

    }


    /**
     * Skaalaa ikkunan sislln annettujen reuna-arvojen mukaisesti, 
     * esimerkiksi:
     * <pre>
     *     window.scale(-2.5, -2.2, 2.1, 2.7);
     * </pre>
     * @param left sislln x-koordinaatin arvo ikkunan vasemmassa reunassa
     * @param bottom sislln y-koordinaatin arvo ikkunan alareunassa 
     * @param right sislln x-koordinaatin arvo ikkunan oikeassa reunassa
     * @param top sislln y-koordinaatin arvo ikkunan ylreunassa
     */
    public void scale(double left, double bottom, double right, double top) {
        panel.scale(left, bottom, right, top);
        redraw();
    }


    /**
     * Kiert ikkunan sislt annetun akselin ympri, esimerkiksi:
     * <pre>
     *     window.rotate(Axis.X, 45.2);
     * </pre>
     * @param axis akseli jonka ympri pyritetn (Axis.X/Axis.Y/Axis.Z)
     * @param deg kierrettvn kulman suuruus asteina
     */
    public void rotate(int axis, double deg) {
        panel.rotate(axis, deg);
    }


    /**
     * Liikuttaa ikkunan sislt, esimerkiksi:
     * <pre>
     *     window.move(0, 1.5, -15);
     * </pre>
     * @param dx liikutettava matka x-akselin suunnassa
     * @param dy liikutettava matka y-akselin suunnassa
     * @param dz liikutettava matka z-akselin suunnassa
     */
    public void move(double dx, double dy, double dz) {
        panel.move(dx, dy, dz);
    }


    /**
     * Skaalaa ikkunan sislln annettujen kertoimien mukaan
     * <pre>
     *      window.scale(2, 1, 1); venytt kuviot kaksinkertaiseksi x-suunnassa
     * </pre>
     * @param sx sislln x-suuntainen kerroin
     * @param sy sislln y-suuntainen kerroin
     * @param sz sislln z-suuntainen kerroin
     */
    public void scale(double sx, double sy, double sz) {
        panel.scale(sx, sy, sz);
    }


    @Override
    public Transformable transform(Matrix m) {
        panel.transform(m);
        return this;
    }


    /**
     * Asettaa ikkunan uudeksi muunnosmatriisiksi annetun matriisin
     * @param m uusi muunnosmatriisi
     */
    public void setTransform(Matrix m) {
        panel.setTransform(m);
    }


    @Override
    public void setRotator(Matrix m) {
        panel.setRotator(m);
    }


    @Override
    public Matrix getRotator() {
        return panel.getRotator();
    }


    /**
     * Asettaa ikkunan nyttmn annettua avaruutta
     * @param space nytettv avaruus
     */
    public void setSpace(Space space) {
        panel.setSpace(space);
    }


    /**
     * @return ikkunan kyttm avaruus
     */
    public Space getSpace() {
        return panel.getSpace();
    }


    /**
     * Lis ikkunaan kuvion joka on Drawable-olio ja palauttaa sen,
     * esimerkiksi:
     * <pre>
     *      window.add(new Line(0, 0, 1, 1)).setColor(255, 255, 0);
     * </pre>
     * @param shape listtv kuvio
     * @return listty kuvio
     */
    public Drawable add(Drawable shape) {
        return panel.add(shape);
    }


    /**
     * Antaa Drawable-olion joka muodostuu kaikista 
     * tmn ikkunan nyttmist kuvioista
     * @return kokoelma tmn ikkunan nyttmist kuvioista
     */
    public Drawable getSavedPath() {
        return panel.getSavedPath();
    }


    /**
     * Antaa ikkunan kokonaismuunnosmatriisin
     * @return ikkunan kokonaismuunnosmatriisi
     */
    public Matrix getFullTransform() {
        return panel.getFullTransform();
    }


    /**
     * Poistaa ikkunasta kuvion joka on Drawable-olio
     * @param shape poistettava kuvio
     */
    public void remove(Drawable shape) {
        panel.remove(shape);
    }


    /**
     * Poistaa ikkunasta kaikki kuviot
     *
     */
    public void clear() {
        panel.removeAll();
    }


    /**
     * Piirt ikkunan sislln uudelleen
     */
    public void redraw() {
        repaint();
    }


    /**
     * Tuo ikkunan nkyville nyttn
     */
    public void showWindow() {
        this.setVisible(true);
    }


    /**
     * Tuo ikkunan nkyville nyttn, piirt ikkunan tiedostoon ja sammuttaa ohjelman
     * @param filename 
     */
    public void showWindow(String filename) {
        this.setVisible(true);
        capture(this, filename);
        System.exit(0);
    }


    /**
     * Asettaa isotrooppisuuden plle/pois
     * @param isotrophic isotrooppisuus tosi/eptosi
     */
    public void setIsotrophic(boolean isotrophic) {
        panel.setIsotrophic(isotrophic);
    }


    /**
     * Lis ikkunaan ruudun jonka avulla voi pyritt
     * ikkunan sislt liikuttamalla hiirt ruudun sisss
     * @param touchPadOn ruutu nkyvill tosi/eptosi
     * @param move tuleeko hiiren napin olla pohjassa tosi/eptosi
     */
    public void setTouchPad(boolean touchPadOn, boolean move) {
        if (touchPadOn) {
            if (touchPad == null)
                touchPad = new ControllerFrame(new TouchPad(panel, move), 0,
                        350, 150, 150);
            touchPad.showController();
        } else {
            if (touchPad != null)
                touchPad.hideController();
        }
    }


    /**
     * Lis ikkunaan ruudun jonka avulla voi pyritt
     * ikkunan sislt liikuttamalla hiirt ruudun sisss
     * siten, ett hiiren nappia ei paineta
     * @param touchPadOn ruutu nkyvill tosi/eptosi
     */
    public void setTouchPad(boolean touchPadOn) {
        this.setTouchPad(touchPadOn, false);
    }


    /**
     * Lis ikkunaan jokaiselle akselille kiertopainikkeet joilla
     * ikkunan sislt voi kiert vastaavien akselien ympri
     * @param controlButtonsOn painikkeet nkyvill tosi/eptosi
     */
    public void setControlButtons(boolean controlButtonsOn) {
        if (controlButtonsOn) {
            if (controlButtons == null)
                controlButtons = new ControllerFrame(new ControlButtons(panel),
                        150, 350, 150, 150);
            controlButtons.showController();
        } else {
            if (controlButtons != null)
                controlButtons.hideController();
        }
    }


    /**
     * Palauttaa tmn ikkunan sisltmn paneelin johon
     * kuviot piirretn
     * @return tmn ikkunan paneeli
     */
    public SpacePanel getPanel() {
        return panel;
    }

    /**
     * Luokka ikkunan sulkemiseen
     */
    protected class CloseWindow extends WindowAdapter {

        /**
         * Ksittelij ikkunan sulkemiselle.
         * @param event ei kytss
         */
        @Override
        public void windowClosing(WindowEvent event) {

            System.exit(0); // NOPMD EI ole J2EE ohjelma
        }
    }


    /**
     * Vaihtaa ikkunan muunnosmatriisin annetuksi matriisiksi
     * @param m uusi muunnosmatriisi
     */
    @Override
    public void changeTransform(Matrix m) {
        panel.changeTransform(m);
    }


    /**
     * Antaa ikkunan muunnosmatriisin
     * @return ikkunan muunnosmatriisi
     */
    @Override
    public Matrix getTransform() {
        return panel.getTransform();
    }
}
